#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ntstatus.h>

#include <cstdio>
#include <cstdlib>

#pragma comment(lib, "ntdll")

extern "C" {

NTSTATUS NTAPI NtOpenKey(
  PHANDLE            KeyHandle,
  ACCESS_MASK        DesiredAccess,
  POBJECT_ATTRIBUTES ObjectAttributes
);

}  // extern "C"

NTSTATUS OpenRelativeKey(HKEY hParentKey,
                         PCWCHAR RelativePath,
                         ACCESS_MASK DesiredAccess,
                         PHKEY OutputKey) {
  UNICODE_STRING UnicodeKeyPath;
  RtlInitUnicodeString(&UnicodeKeyPath, RelativePath);

  OBJECT_ATTRIBUTES ObjectAttributes;
  InitializeObjectAttributes(&ObjectAttributes,
                             &UnicodeKeyPath,
                             0,
                             hParentKey,
                             NULL);

  return NtOpenKey((PHANDLE)OutputKey, DesiredAccess, &ObjectAttributes);
}

LONG EnumerateSubKeys(HKEY hKey) {
  for (DWORD i = 0;; i++) {
    WCHAR KeyName[100];
    DWORD cchKeyName = sizeof(KeyName) / sizeof(KeyName[0]);

    LONG st = RegEnumKeyExW(hKey, i, KeyName, &cchKeyName,
                            NULL, NULL, NULL, NULL);

    if (st == ERROR_NO_MORE_ITEMS) {
      break;
    } else if (st != ERROR_SUCCESS) {
      wprintf(L"RegEnumKeyExW failed with error %d\n", st);
      return st;
    }

    wprintf(L"%d. %s\n", i + 1, KeyName);
  }

  return ERROR_SUCCESS;
}

int wmain(int argc, wchar_t **argv) {
  NTSTATUS Status;
  LONG st;

  if (argc != 2) {
    wprintf(L"Usage: %s <raw registry path>\n", argv[0]);
    return 1;
  }

  if (_wcsnicmp(argv[1], L"\\Registry\\", 10) != 0) {
    wprintf(L"Registry path must start with \\Registry\\\n");
    return 1;
  }

  HKEY hRootKey;
  Status = OpenRelativeKey(NULL, L"\\Registry", KEY_READ, &hRootKey);

  if (!NT_SUCCESS(Status)) {
    wprintf(L"Opening root key failed with error %.8x\n", Status);
    return 1;
  }

  HKEY hKey;
  Status = OpenRelativeKey(hRootKey, &argv[1][10], KEY_READ, &hKey);

  if (!NT_SUCCESS(Status)) {
    wprintf(L"Opening relative key failed with error %.8x\n", Status);
    return 1;
  }

  st = EnumerateSubKeys(hKey);
  if (st != ERROR_SUCCESS) {
    return 1;
  }

  return 0;
}
