#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ntstatus.h>

#include <cstdio>
#include <cstdlib>

#pragma comment(lib, "ntdll")

extern "C" {

NTSTATUS NTAPI NtOpenKey(
  PHANDLE KeyHandle,
  ACCESS_MASK DesiredAccess,
  POBJECT_ATTRIBUTES ObjectAttributes
);

}  // extern "C"

NTSTATUS OpenRelativeKey(HKEY hParentKey,
                         PCWCHAR RelativePath,
                         ACCESS_MASK DesiredAccess,
                         PHKEY OutputKey) {
  UNICODE_STRING UnicodeKeyPath;
  RtlInitUnicodeString(&UnicodeKeyPath, RelativePath);

  OBJECT_ATTRIBUTES ObjectAttributes;
  InitializeObjectAttributes(&ObjectAttributes,
                             &UnicodeKeyPath,
                             0,
                             hParentKey,
                             NULL);

  return NtOpenKey((PHANDLE)OutputKey, DesiredAccess, &ObjectAttributes);
}

LONG EnumerateSubKeys(HKEY hKey) {
  for (DWORD i = 0;; i++) {
    WCHAR KeyName[100];
    DWORD cchKeyName = sizeof(KeyName) / sizeof(KeyName[0]);

    LONG st = RegEnumKeyExW(hKey, i, KeyName, &cchKeyName,
                            NULL, NULL, NULL, NULL);

    if (st == ERROR_NO_MORE_ITEMS) {
      break;
    } else if (st != ERROR_SUCCESS) {
      wprintf(L"RegEnumKeyExW failed with error %d\n", st);
      return st;
    }

    wprintf(L"%d. %s\n", i + 1, KeyName);
  }

  return ERROR_SUCCESS;
}

int wmain() {
  NTSTATUS Status;
  LONG st;
  
  HKEY hKeyRegistry;
  Status = OpenRelativeKey(NULL, L"\\Registry", KEY_READ, &hKeyRegistry);

  if (!NT_SUCCESS(Status)) {
    wprintf(L"Opening key #1 failed with error %.8x\n", Status);
    return 1;
  }

  CONST ULONG KeyPathLength = 32758;
  CONST ULONG KeyPathBufferLength = KeyPathLength * sizeof(WCHAR);

  PWCHAR KeyPath = (PWCHAR)malloc(KeyPathBufferLength);
  wcscpy_s(KeyPath, KeyPathLength, L"User");

  for (ULONG i = wcslen(KeyPath); i < KeyPathLength - 1; i++) {
    KeyPath[i] = L'\\';
  }
  KeyPath[KeyPathLength - 1] = L'\0';

  HKEY hKeyUsers;
  Status = OpenRelativeKey(hKeyRegistry, KeyPath, KEY_READ, &hKeyUsers);

  if (!NT_SUCCESS(Status)) {
    wprintf(L"Opening key #2 failed with error %.8x\n", Status);
    return 1;
  }

  wprintf(L"Subkeys of containerized HKEY_USERS:\n");
  st = EnumerateSubKeys(HKEY_USERS);

  if (st != ERROR_SUCCESS) {
    return 1;
  }

  wprintf(L"\nSubkeys of host HKEY_USERS:\n");
  st = EnumerateSubKeys(hKeyUsers);

  if (st != ERROR_SUCCESS) {
    return 1;
  }

  return 0;
}
